



cd tests
IF %ERRORLEVEL% NEQ 0 exit /B 1
set SKIP_OMP=true
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest -vv test_imagecodecs.py -k "not test_image_roundtrips[jpeg2k-uint16-rgb-out-bytearray-None]"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
